import os

from neuralogic.utils.data.dataset import Dataset, Data
from neuralogic.core.settings import Settings

base_path = os.path.join(os.path.abspath(os.path.dirname(__file__)), "datasets")


def XOR(settings: Settings):
    from neuralogic.core.template import Template

    template = Template(settings, template_file=os.path.join(base_path, "simple", "xor", "naive", "template.txt"))
    dataset = Dataset(examples_file=os.path.join(base_path, "simple", "xor", "naive", "trainExamples.txt"))

    return template, dataset


def Trains(settings: Settings):
    from neuralogic.core.template import Template

    template = Template(settings, template_file=os.path.join(base_path, "simple", "trains", "template.txt"))
    dataset = Dataset(
        examples_file=os.path.join(base_path, "simple", "trains", "examples.txt"),
        queries_file=os.path.join(base_path, "simple", "trains", "queries.txt"),
    )

    return template, dataset


def XOR_Vectorized(settings: Settings):
    from neuralogic.core.template import Template

    template = Template(settings, template_file=os.path.join(base_path, "simple", "xor", "vectorized", "template.txt"))
    dataset = Dataset(examples_file=os.path.join(base_path, "simple", "xor", "vectorized", "trainExamples.txt"))

    return template, dataset


def Mutagenesis(settings: Settings):
    from neuralogic.core.template import Template

    template = Template(
        settings,
        template_file=os.path.join(base_path, "molecules", "mutagenesis", "templates", "template.txt"),
    )

    dataset = Dataset(
        examples_file=os.path.join(base_path, "molecules", "mutagenesis", "examples.txt"),
        queries_file=os.path.join(base_path, "molecules", "mutagenesis", "queries.txt"),
    )

    return template, dataset



from neuralogic import get_neuralogic, get_gateway
from py4j.java_gateway import set_field, get_field
from enum import Enum


class Optimizer(str, Enum):
    ADAM = "ADAM"
    SGD = "SGD"


class ErrorFunction(str, Enum):
    SQUARED_DIFF = "SQUARED_DIFF"
    # ABS_DIFF = "ABS_DIFF"
    CROSSENTROPY = "CROSSENTROPY"


class Activation(Enum):
    SIGMOID = "sigmoid"
    TANH = "tanh"
    SIGNUM = "signum"
    RELU = "relu"
    IDENTITY = "identity"
    LUKASIEWICZ = "lukasiewicz"
    SOFTMAX = "softmax"
    SPARSEMAX = "sparsemax"


class Aggregation(Enum):
    SUM = "sum"
    MAX = "max"
    AVG = "avg"


class Initializer(Enum):
    UNIFORM = "uniform"
    NORMAL = "normal"
    CONSTANT = "constant"
    LONGTAIL = "longtail"
    GLOROT = "glorot"
    HE = "he"


class Settings:
    def __init__(
        self,
        learning_rate: float = None,
        optimizer: Optimizer = None,
        epochs: int = None,
        error_function: ErrorFunction = None,
        initializer: Initializer = None,
        initializer_const: float = None,
        initializer_uniform_scale: float = None,
    ):
        self.namespace = get_neuralogic().cz.cvut.fel.ida.setup
        self.settings = self.namespace.Settings()

        if learning_rate is not None:
            self.learning_rate = learning_rate

        if optimizer is not None:
            self.optimizer = optimizer

        if initializer is not None:
            self.initializer = initializer

        if initializer_const is not None:
            self.initializer_const = initializer_const

        if initializer_uniform_scale is not None:
            self.initializer_uniform_scale = initializer_uniform_scale

        if epochs is not None:
            self.epochs = epochs

        self.error_function = ErrorFunction.SQUARED_DIFF if error_function is None else error_function
        set_field(self.settings, "debugExporting", False)
        set_field(self.settings, "isoValueCompression", True)
        set_field(self.settings, "exportBlocks", get_gateway().new_array(get_gateway().jvm.java.lang.String, 0))
        self.settings.infer()

    @property
    def iso_value_compression(self) -> bool:
        return get_field(self.settings, "isoValueCompression")

    @iso_value_compression.setter
    def iso_value_compression(self, iso_value_compression: bool):
        set_field(self.settings, "isoValueCompression", iso_value_compression)

    @property
    def seed(self) -> int:
        return get_field(self.settings, "seed")

    @seed.setter
    def seed(self, seed: int):
        set_field(self.settings, "seed", seed)

    @property
    def learning_rate(self) -> float:
        return get_field(self.settings, "initLearningRate")

    @learning_rate.setter
    def learning_rate(self, learning_rate: float):
        set_field(self.settings, "initLearningRate", learning_rate)

    @property
    def optimizer(self):
        return self.settings.getOptimizer()

    @optimizer.setter
    def optimizer(self, optimizer: Optimizer):
        if optimizer == Optimizer.SGD:
            java_optimizer = self.namespace.Settings.OptimizerSet.SGD
        elif optimizer == Optimizer.ADAM:
            java_optimizer = self.namespace.Settings.OptimizerSet.ADAM
        else:
            raise NotImplementedError
        self.settings.setOptimizer(java_optimizer)

    @property
    def initializer_const(self):
        return get_field(self.settings, "constantInitValue")

    @initializer_const.setter
    def initializer_const(self, value: float):
        set_field(self.settings, "constantInitValue", value)

    @property
    def initializer_uniform_scale(self):
        return get_field(self.settings, "randomInitScale")

    @initializer_uniform_scale.setter
    def initializer_uniform_scale(self, value: float):
        set_field(self.settings, "randomInitScale", value)

    @property
    def error_function(self):
        return self.settings.errorFunction

    @error_function.setter
    def error_function(self, error_function: ErrorFunction):
        if error_function == ErrorFunction.SQUARED_DIFF:
            java_error_function = self.namespace.Settings.ErrorFcn.SQUARED_DIFF
        # elif error_function == ErrorFunction.ABS_DIFF:
        #     java_error_function = self.namespace.Settings.ErrorFcn.ABS_DIFF
        elif error_function == ErrorFunction.CROSSENTROPY:
            java_error_function = self.namespace.Settings.ErrorFcn.CROSSENTROPY
        else:
            raise NotImplementedError
        self.settings.errorFunction = java_error_function

    @property
    def epochs(self) -> int:
        return get_field(self.settings, "maxCumEpochCount")

    @epochs.setter
    def epochs(self, epochs: int):
        set_field(self.settings, "maxCumEpochCount", epochs)

    @property
    def initializer(self):
        initializer = get_field(self.settings, "initializer")

        if str(initializer) != "SIMPLE":
            return initializer
        return get_field(self.settings, "initDistribution")

    @initializer.setter
    def initializer(self, initializer: Initializer):
        if initializer == Initializer.HE:
            set_field(self.settings, "initializer", self.namespace.Settings.InitSet.HE)
            return
        if initializer == Initializer.GLOROT:
            set_field(self.settings, "initializer", self.namespace.Settings.InitSet.GLOROT)
            return

        if initializer == Initializer.NORMAL:
            init_dist = self.namespace.Settings.InitDistribution.NORMAL
        elif initializer == Initializer.UNIFORM:
            init_dist = self.namespace.Settings.InitDistribution.UNIFORM
        elif initializer == Initializer.CONSTANT:
            init_dist = self.namespace.Settings.InitDistribution.CONSTANT
        elif initializer == Initializer.LONGTAIL:
            init_dist = self.namespace.Settings.InitDistribution.LONGTAIL
        else:
            raise NotImplementedError
        set_field(self.settings, "initDistribution", init_dist)
        set_field(self.settings, "initializer", self.namespace.Settings.InitSet.SIMPLE)

    @property
    def debug_exporting(self) -> bool:
        return get_field(self.settings, "debugExporting")

    @debug_exporting.setter
    def debug_exporting(self, debug_export: bool):
        set_field(self.settings, "debugExporting", debug_export)

    @property
    def default_fact_value(self) -> float:
        return get_field(self.settings, "defaultFactValue")

    @default_fact_value.setter
    def default_fact_value(self, value: float):
        set_field(self.settings, "defaultFactValue", value)

    def to_json(self) -> str:
        return self.settings.exportToJson()



import json
from collections import Sized
from typing import Optional, Dict
from py4j.java_gateway import get_field
from py4j.java_collections import SetConverter

from neuralogic import get_neuralogic, get_gateway
from neuralogic.nn.base import AbstractNeuraLogic
from neuralogic.core.settings import Settings


class Loss:
    def __init__(self, loss):
        self.loss = loss

    def backward(self):
        self.loss.backward()

    def value(self) -> float:
        return get_field(self.loss.getError(), "value")

    def output(self) -> float:
        return get_field(self.loss.getOutput(), "value")

    def target(self) -> float:
        return get_field(self.loss.getTarget(), "value")


class NeuraLogic(AbstractNeuraLogic):
    class HookHandler:
        def __init__(self, module: "NeuraLogic"):
            self.module = module

        def handleHook(self, hook, value):
            self.module.run_hook(hook, json.loads(value))

        class Java:
            implements = ["cz.cvut.fel.ida.neural.networks.computation.iteration.actions.PythonHookHandler"]

    def __init__(self, model, template, settings: Optional[Settings] = None):
        super().__init__(template)
        self.namespace = get_neuralogic().cz.cvut.fel.ida.neural.networks.computation.training.strategies
        self.do_train = True
        self.need_sync = False

        if settings is None:
            settings = Settings()

        self.settings = settings
        self.neural_model = model
        self.strategy = self.namespace.PythonTrainingStrategy(settings.settings, model)
        self.samples_len = 0

        self.hook_handler = NeuraLogic.HookHandler(self)
        self.reset_parameters()

    def reset_parameters(self):
        self.strategy.resetParameters()

    def train(self):
        self.do_train = True

    def test(self):
        self.do_train = False

    def set_training_samples(self, samples):
        self.samples_len = len(samples)
        self.strategy.setSamples(samples)

    def __call__(self, samples=None, train: bool = None, auto_backprop: bool = False, epochs: int = 1):
        self.hooks_set = len(self.hooks) != 0

        if self.hooks_set:
            self.strategy.setHooks(
                SetConverter().convert(set(self.hooks.keys()), get_gateway()._gateway_client), self.hook_handler
            )

        if train is not None:
            self.do_train = train

        if samples is None:
            return self.strategy.learnSamples(epochs), self.samples_len

        if not isinstance(samples, Sized):
            if self.do_train:
                if auto_backprop:
                    return self.strategy.learnSample(samples), 1
            result = self.strategy.evaluateSample(samples)
            return Loss(result)

        if self.do_train:
            return self.strategy.learnSamples(samples, epochs), len(samples)

        results = self.strategy.evaluateSamples(samples)
        return [(get_field(result.getTarget(), "value"), get_field(result.getOutput(), "value")) for result in results]

    def state_dict(self) -> Dict:
        weights = self.neural_model.getAllWeights()
        weights_dict = {}

        for weight in weights:
            if get_field(weight, "isLearnable"):
                value = get_field(weight, "value")

                size = list(value.size())

                if len(size) == 0 or size[0] == 0:
                    weights_dict[get_field(weight, "index")] = get_field(value, "value")
                elif len(size) == 1 or size[0] == 1 or size[1] == 1:
                    weights_dict[get_field(weight, "index")] = list(get_field(value, "values"))
                else:
                    weights_dict[get_field(weight, "index")] = json.loads(value.toString())
        return {
            "weights": weights_dict,
        }

    def load_state_dict(self, state_dict: Dict):
        self.sync_template(state_dict, self.neural_model.getAllWeights())



from typing import Optional, List, Union, Sized

from neuralogic.core.constructs.atom import BaseAtom, WeightedAtom
from neuralogic.core.constructs.rule import Rule


DatasetEntries = Union[BaseAtom, WeightedAtom, Rule]


class Data:
    """
    Stores a learning example in the form of a tensor numeric representation instead of a rule based representation.
    """
    def __init__(
        self,
        x: Sized = None,
        edge_index: Sized = None,
        edge_attr: Optional[Sized] = None,
        y_mask: Sized = None,
        y: Sized = None,
    ):
        self.x = x
        self.edge_index = edge_index
        self.edge_attr = edge_attr
        self.y = y
        self.y_mask = y_mask

    @staticmethod
    def from_pyg(data) -> List["Data"]:
        """
        Converts a PyTorch Geometric Data instance into a list of PyNeuraLogic :py:class:`~Data` instances.
        The conversion supports :code:`train_mask`, :code:`test_mask` and :code:`val_mask` attributes -
        for each mask the conversion yields a new data instance.

        :param data: The PyTorch Geometric Data instance
        :return: The list of PyNeuraLogic Data instances
        """
        data_list = []

        if hasattr(data, "train_mask"):
            data_list.append(Data(data.x, data.edge_index, data.edge_attr, data.train_mask, data.y))
        if hasattr(data, "test_mask"):
            data_list.append(Data(data.x, data.edge_index, data.edge_attr, data.test_mask, data.y))
        if hasattr(data, "val_mask"):
            data_list.append(Data(data.x, data.edge_index, data.edge_attr, data.val_mask, data.y))
        if len(data_list) == 0:
            data_list.append(Data(data.x, data.edge_index, data.edge_attr, None, data.y))

        return data_list


class Dataset:
    def __init__(
        self,
        *,
        data: Optional[List[Data]] = None,
        examples_file: Optional[str] = None,
        queries_file: Optional[str] = None,
    ):
        self.file_sources = False

        if examples_file is not None or queries_file is not None:
            if data is not None:
                raise Exception
            self.file_sources = True

        self.data = data
        self.examples_file = examples_file
        self.queries_file = queries_file

        self.examples: List[DatasetEntries] = []
        self.queries: List[DatasetEntries] = []

    def __len__(self):
        if self.data is not None:
            return len(self.data)
        if self.file_sources:
            return -1
        return max(len(self.examples), len(self.queries))

    def __getitem__(self, item):
        pass

    def add_example(self, example):
        self.add_examples([example])

    def add_examples(self, examples: List):
        if self.file_sources or self.data is not None:
            raise Exception
        self.examples.extend(examples)

    def set_examples(self, examples: List):
        if self.file_sources or self.data is not None:
            raise Exception
        self.examples = examples

    def add_query(self, query):
        self.add_queries([query])

    def add_queries(self, queries: List):
        if self.file_sources or self.data is not None:
            raise Exception
        self.queries.extend(queries)

    def set_queries(self, queries: List):
        if self.file_sources or self.data is not None:
            raise Exception
        self.queries = queries



from typing import Union, Optional
from pathlib import Path
from contextlib import contextmanager
import os
from py4j.java_gateway import JavaGateway, JVMView, CallbackServerParameters
from py4j.protocol import unescape_new_line

neuralogic: Optional[JVMView] = None
gateway: Optional[JavaGateway] = None

std_out = None
std_err = None
port = os.environ.get("NEURALOGIC_PORT", 25333)


def set_java_home(java_home: Union[Path, str]):
    """
    Set java home environment variable

    :param java_home:
    :return:
    """
    os.environ["JAVA_HOME"] = str(java_home)


def set_gateway_port(gateway_port: int):
    global port
    port = gateway_port


def set_std_out(out):
    global std_out
    std_out = out


def set_std_err(out):
    global std_err
    std_err = out


def initialize(gateway_port: Optional[int] = None, die_on_exit: bool = True):
    """
    Initialize the gateway for the Java process. Has to be reinitialized if some settings (JAVA_HOME) changes

    :param gateway_port:
    :param die_on_exit:
    :return:
    """
    global gateway, neuralogic

    if gateway_port is None:
        gateway_port = port

    if gateway is not None:
        try:
            gateway.shutdown()
            gateway = None
        except Exception:
            pass

    for try_port in range(gateway_port, gateway_port + 10):
        try:
            gateway = JavaGateway.launch_gateway(
                classpath=os.environ["CLASSPATH"],
                redirect_stdout=std_out,
                redirect_stderr=std_err,
                die_on_exit=die_on_exit,
                daemonize_redirect=True,
            )

            raw_token = unescape_new_line(gateway.gateway_parameters.auth_token)

            params = CallbackServerParameters(
                eager_load=False,
                auth_token=raw_token,
                daemonize_connections=True,
                daemonize=True,
                port=try_port,
            )

            gateway.start_callback_server(params)
            neuralogic = gateway.jvm

            return
        except Exception as e:
            if gateway is not None:
                gateway.shutdown()
                gateway = None

    if gateway is not None:
        gateway.shutdown()
        gateway = None

    raise Exception(
        f"Cannot find two free ports in the range <{gateway_port}-{gateway_port + 10 + 1}>.\n"
        "Please specify different port in env var 'NEURALOGIC_PORT' or via neuralogic.set_gateway_port."
    )


def shutdown():
    if gateway is not None:
        gateway.shutdown()


def get_gateway() -> JavaGateway:
    """
    Get the gateway for the Java process

    :return:
    """
    if gateway is None:
        initialize()
    return gateway


def get_neuralogic() -> JVMView:
    """
    Get the jvm view of the Java process

    :return:
    """
    if neuralogic is None:
        initialize()
    return neuralogic


@contextmanager
def neuralogic_jvm():
    if gateway is None:
        initialize()
    yield
    if gateway is not None:
        shutdown()


os.environ["CLASSPATH"] = os.path.join(os.path.abspath(os.path.dirname(__file__)), "jar", "NeuraLogic.jar")



from typing import Optional, Dict, Union

import torch.nn.functional as F
import torch

from neuralogic.nn.base import AbstractEvaluator

from neuralogic.core import Template, BuiltDataset
from neuralogic.core.settings import Settings, Optimizer, ErrorFunction
from neuralogic.core.builder import Backend
from neuralogic.utils.data import Dataset


class TorchEvaluator(AbstractEvaluator):
    trainers = {
        Optimizer.SGD: lambda param, rate: torch.optim.SGD(param, lr=rate),
        Optimizer.ADAM: lambda param, rate: torch.optim.Adam(param, lr=rate),
    }

    error_functions = {ErrorFunction.SQUARED_DIFF: F.mse_loss, ErrorFunction.CROSSENTROPY: F.cross_entropy}

    def __init__(self, template: Template, settings: Settings):
        super().__init__(Backend.PYG, template, settings)

    def train(self, dataset: Optional[Union[Dataset, BuiltDataset]] = None, *, generator: bool = True):
        # dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        epochs = self.settings.epochs
        error_function = ErrorFunction[str(self.settings.error_function)]
        optimizer = Optimizer[str(self.settings.optimizer)]

        if optimizer not in TorchEvaluator.trainers:
            raise NotImplementedError
        if error_function not in TorchEvaluator.error_functions:
            raise NotImplementedError

        trainer = TorchEvaluator.trainers[optimizer](
            self.neuralogic_model.module_list.parameters(),
            self.settings.learning_rate,
        )
        error_function = TorchEvaluator.error_functions[error_function]

        def _train():
            for _ in range(epochs):
                seen_instances = 0
                total_loss = 0

                for data in dataset.data:
                    self.neuralogic_model.train()
                    trainer.zero_grad()

                    out = self.neuralogic_model(x=data.x, edge_index=data.edge_index)
                    loss = error_function(out[data.y_mask], data.y[data.y_mask])
                    loss.backward()
                    trainer.step()

                    seen_instances += 1
                    total_loss += float(loss)
                yield total_loss, seen_instances

        if generator:
            return _train()

        stats = 0, 0
        for stats in _train():
            pass
        return stats

    def test(self, dataset: Optional[Union[Dataset, BuiltDataset]] = None, *, generator: bool = True):
        self.neuralogic_model.train(mode=False)

        # dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        def _test():
            for data in dataset.data:
                self.neuralogic_model.train(mode=False)
                out = self.neuralogic_model(x=data.x, edge_index=data.edge_index)
                results = (out[data.y_mask], data.y[data.y_mask])

                yield results

        if generator:
            return _test()
        return list(_test())



# Generated from /home/lukas/Workspace/prcvut/pyneuralogic/neuralogic/grammar/Neuralogic.g4 by ANTLR 4.8
from antlr4 import *

if __name__ is not None and "." in __name__:
    from .NeuralogicParser import NeuralogicParser
else:
    from NeuralogicParser import NeuralogicParser

# This class defines a complete listener for a parse tree produced by NeuralogicParser.
class NeuralogicListener(ParseTreeListener):

    # Enter a parse tree produced by NeuralogicParser#templateFile.
    def enterTemplateFile(self, ctx: NeuralogicParser.TemplateFileContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#templateFile.
    def exitTemplateFile(self, ctx: NeuralogicParser.TemplateFileContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#templateLine.
    def enterTemplateLine(self, ctx: NeuralogicParser.TemplateLineContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#templateLine.
    def exitTemplateLine(self, ctx: NeuralogicParser.TemplateLineContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#examplesFile.
    def enterExamplesFile(self, ctx: NeuralogicParser.ExamplesFileContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#examplesFile.
    def exitExamplesFile(self, ctx: NeuralogicParser.ExamplesFileContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#liftedExample.
    def enterLiftedExample(self, ctx: NeuralogicParser.LiftedExampleContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#liftedExample.
    def exitLiftedExample(self, ctx: NeuralogicParser.LiftedExampleContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#label.
    def enterLabel(self, ctx: NeuralogicParser.LabelContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#label.
    def exitLabel(self, ctx: NeuralogicParser.LabelContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#queriesFile.
    def enterQueriesFile(self, ctx: NeuralogicParser.QueriesFileContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#queriesFile.
    def exitQueriesFile(self, ctx: NeuralogicParser.QueriesFileContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#fact.
    def enterFact(self, ctx: NeuralogicParser.FactContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#fact.
    def exitFact(self, ctx: NeuralogicParser.FactContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#atom.
    def enterAtom(self, ctx: NeuralogicParser.AtomContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#atom.
    def exitAtom(self, ctx: NeuralogicParser.AtomContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#termList.
    def enterTermList(self, ctx: NeuralogicParser.TermListContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#termList.
    def exitTermList(self, ctx: NeuralogicParser.TermListContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#term.
    def enterTerm(self, ctx: NeuralogicParser.TermContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#term.
    def exitTerm(self, ctx: NeuralogicParser.TermContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#variable.
    def enterVariable(self, ctx: NeuralogicParser.VariableContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#variable.
    def exitVariable(self, ctx: NeuralogicParser.VariableContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#constant.
    def enterConstant(self, ctx: NeuralogicParser.ConstantContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#constant.
    def exitConstant(self, ctx: NeuralogicParser.ConstantContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#predicate.
    def enterPredicate(self, ctx: NeuralogicParser.PredicateContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#predicate.
    def exitPredicate(self, ctx: NeuralogicParser.PredicateContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#conjunction.
    def enterConjunction(self, ctx: NeuralogicParser.ConjunctionContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#conjunction.
    def exitConjunction(self, ctx: NeuralogicParser.ConjunctionContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#metadataVal.
    def enterMetadataVal(self, ctx: NeuralogicParser.MetadataValContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#metadataVal.
    def exitMetadataVal(self, ctx: NeuralogicParser.MetadataValContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#metadataList.
    def enterMetadataList(self, ctx: NeuralogicParser.MetadataListContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#metadataList.
    def exitMetadataList(self, ctx: NeuralogicParser.MetadataListContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#lrnnRule.
    def enterLrnnRule(self, ctx: NeuralogicParser.LrnnRuleContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#lrnnRule.
    def exitLrnnRule(self, ctx: NeuralogicParser.LrnnRuleContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#predicateOffset.
    def enterPredicateOffset(self, ctx: NeuralogicParser.PredicateOffsetContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#predicateOffset.
    def exitPredicateOffset(self, ctx: NeuralogicParser.PredicateOffsetContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#predicateMetadata.
    def enterPredicateMetadata(self, ctx: NeuralogicParser.PredicateMetadataContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#predicateMetadata.
    def exitPredicateMetadata(self, ctx: NeuralogicParser.PredicateMetadataContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#weightMetadata.
    def enterWeightMetadata(self, ctx: NeuralogicParser.WeightMetadataContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#weightMetadata.
    def exitWeightMetadata(self, ctx: NeuralogicParser.WeightMetadataContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#templateMetadata.
    def enterTemplateMetadata(self, ctx: NeuralogicParser.TemplateMetadataContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#templateMetadata.
    def exitTemplateMetadata(self, ctx: NeuralogicParser.TemplateMetadataContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#weight.
    def enterWeight(self, ctx: NeuralogicParser.WeightContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#weight.
    def exitWeight(self, ctx: NeuralogicParser.WeightContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#fixedValue.
    def enterFixedValue(self, ctx: NeuralogicParser.FixedValueContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#fixedValue.
    def exitFixedValue(self, ctx: NeuralogicParser.FixedValueContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#offset.
    def enterOffset(self, ctx: NeuralogicParser.OffsetContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#offset.
    def exitOffset(self, ctx: NeuralogicParser.OffsetContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#value.
    def enterValue(self, ctx: NeuralogicParser.ValueContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#value.
    def exitValue(self, ctx: NeuralogicParser.ValueContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#number.
    def enterNumber(self, ctx: NeuralogicParser.NumberContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#number.
    def exitNumber(self, ctx: NeuralogicParser.NumberContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#vector.
    def enterVector(self, ctx: NeuralogicParser.VectorContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#vector.
    def exitVector(self, ctx: NeuralogicParser.VectorContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#matrix.
    def enterMatrix(self, ctx: NeuralogicParser.MatrixContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#matrix.
    def exitMatrix(self, ctx: NeuralogicParser.MatrixContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#dimensions.
    def enterDimensions(self, ctx: NeuralogicParser.DimensionsContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#dimensions.
    def exitDimensions(self, ctx: NeuralogicParser.DimensionsContext):
        pass

    # Enter a parse tree produced by NeuralogicParser#negation.
    def enterNegation(self, ctx: NeuralogicParser.NegationContext):
        pass

    # Exit a parse tree produced by NeuralogicParser#negation.
    def exitNegation(self, ctx: NeuralogicParser.NegationContext):
        pass


del NeuralogicParser



from neuralogic.core.settings import Activation, Aggregation


class Metadata:
    def __init__(
        self, offset=None, learnable: bool = None, activation: Activation = None, aggregation: Aggregation = None
    ):
        self.offset = offset
        self.learnable = learnable
        self.activation = activation
        self.aggregation = aggregation

    def __str__(self):
        metadata_list = []
        if self.offset is not None:
            metadata_list.append(f"offset={self.offset}")
        if self.learnable is not None:
            metadata_list.append(f"learnable={str(self.learnable).lower()}")
        if self.activation is not None:
            metadata_list.append(f"activation={self.activation.value}")
        if self.aggregation is not None:
            metadata_list.append(f"aggregation={self.aggregation.value}")
        return f"[{', '.join(metadata_list)}]"



from typing import Dict
from neuralogic.core.constructs.predicate import Predicate
from neuralogic.core.constructs import atom


class AtomFactory:
    class Predicate:
        predicates: Dict[str, Predicate] = {}

        def __init__(self, private=False, special=False):
            self.is_private = private
            self.is_special = special

        @property
        def special(self) -> "AtomFactory.Predicate":
            return AtomFactory.Predicate(self.is_private, True)

        @property
        def private(self) -> "AtomFactory.Predicate":
            return AtomFactory.Predicate(True, self.is_special)

        @staticmethod
        def get_predicate(name, arity, private, special) -> Predicate:
            key = f"{name}/{arity}"
            if key not in AtomFactory.Predicate.predicates:
                AtomFactory.Predicate.predicates[key] = Predicate(name, arity, private, special)
            return AtomFactory.Predicate.predicates[key]

        def __getattr__(self, item):
            return atom.BaseAtom(AtomFactory.Predicate.get_predicate(item, 0, self.is_private, self.is_special))

    def __init__(self):
        self.instances: Dict[str, Dict[int, atom.BaseAtom]] = {}

        self.special = AtomFactory.Predicate(special=True)
        self.private = AtomFactory.Predicate(private=True)

    def get(self, name: str) -> atom.BaseAtom:
        return atom.BaseAtom(AtomFactory.Predicate.get_predicate(name, 0, False, False))

    def __getattr__(self, item) -> atom.BaseAtom:
        return atom.BaseAtom(AtomFactory.Predicate.get_predicate(item, 0, False, False))


class VariableFactory:
    def __getattr__(self, item) -> str:
        return item.upper()


class TermFactory:
    def __getattr__(self, item) -> str:
        return item.lower()


Var = VariableFactory()
Relation = AtomFactory()
Term = TermFactory()


V = Var
T = Term
R = Relation


from typing import Optional, Dict, Union

from py4j.java_collections import ListConverter

from neuralogic import get_gateway
from neuralogic.core.builder import Backend
from neuralogic.core import Template, BuiltDataset
from neuralogic.nn.base import AbstractEvaluator
from neuralogic.core.settings import Settings
from neuralogic.utils.data import Dataset


class JavaEvaluator(AbstractEvaluator):
    def __init__(
        self,
        problem: Optional[Template],
        settings: Settings,
    ):
        super().__init__(Backend.JAVA, problem, settings)

    def set_dataset(self, dataset: Union[Dataset, BuiltDataset]):
        super().set_dataset(dataset)
        self.neuralogic_model.set_training_samples(self.dataset.samples)

    def reset_dataset(self, dataset):
        if dataset is None:
            empty_list = ListConverter().convert([], get_gateway()._gateway_client)
            self.neuralogic_model.set_training_samples(empty_list)
        else:
            self.neuralogic_model.set_training_samples(dataset.samples)
        self.dataset = dataset

    def train(
        self, dataset: Optional[Union[Dataset, BuiltDataset]] = None, *, generator: bool = True, epochs: int = None
    ):
        old_dataset = None

        if dataset is not None:
            old_dataset = self.dataset
            self.set_dataset(dataset)

        if epochs is None:
            epochs = self.settings.epochs

        def _train():
            for _ in range(epochs):
                result = self.neuralogic_model(None, True)
                yield result
            if dataset is not None:
                self.reset_dataset(old_dataset)

        if generator:
            return _train()

        results = self.neuralogic_model(None, True, epochs=epochs)
        if dataset is not None:
            self.reset_dataset(old_dataset)

        return results

    def test(self, dataset: Optional[Union[Dataset, BuiltDataset]] = None, *, generator: bool = True):
        dataset = self.dataset if dataset is None else self.build_dataset(dataset)

        def _test():
            for sample in dataset.samples:
                result = self.neuralogic_model(sample, False)
                yield result.target(), result.output()

        if generator:
            return _test()
        return self.neuralogic_model(dataset.samples, False)

    def state_dict(self) -> Dict:
        return self.neuralogic_model.state_dict()

    def load_state_dict(self, state_dict: Dict):
        self.neuralogic_model.load_state_dict(state_dict)



from typing import List, Optional, Tuple

from neuralogic.core.constructs.rule import Rule, Metadata
from neuralogic.core.constructs.atom import WeightedAtom
from neuralogic.utils.templates import GCNConv, SAGEConv, GINConv, Embedding, GlobalPooling, TemplateList


def translate_rules(rules):
    rule_context = {}

    for rule in rules:
        head = rule.head
        head_name = str(head.predicate)

        if head_name not in rule_context:
            rule_context[head_name] = {
                "dependent_on": set(),
                "rules": [],
            }

        rule_context[head_name]["rules"].append(rule)

        for atom in rule.body:
            rule_context[head_name]["dependent_on"].add(str(atom.predicate))

    last_layer = None

    for rule_group in rule_context.keys():
        for context in rule_context.values():
            if rule_group in context["dependent_on"]:
                break
        else:
            last_layer = rule_group
            break

    if last_layer is None:
        return None
    return get_template_list(rule_context, last_layer)


def get_template_list(rule_context, last_layer):
    module_list = []

    current_layer = rule_context[last_layer]

    while True:
        rules, metadata = divide_by_type(current_layer["rules"])
        module_info = get_module(rules)

        if module_info is None:
            return None

        module, dependency = module_info

        module_list.append(module)

        if dependency not in rule_context:
            return TemplateList(list(reversed(module_list)))
        current_layer = rule_context[dependency]


def get_module(rules: List[Rule]):
    if len(rules) == 1:
        rule = rules[0]
        for fn in [get_gcn, get_embedding]:
            module_info = fn(rule)
            if module_info is not None:
                return module_info
        return None
    for fn in [get_gin, get_sage, get_pooling]:
        module_info = fn(rules)
        if module_info is not None:
            return module_info
    return None


def divide_by_type(rules):
    metadata = []
    rules_only = []

    for rule in rules:
        if isinstance(rule, Rule):
            rules_only.append(rule)
        elif isinstance(rule, Metadata):
            metadata.append(rule)
    return rules_only, metadata


def get_in_out_from_atom(atom):
    in_channels, out_channels = 1, 1

    if isinstance(atom, WeightedAtom) and isinstance(atom.weight, tuple):
        if len(atom.weight) == 2:
            out_channels, in_channels = atom.weight
        if len(atom.weight) == 1:
            (out_channels,) = atom.weight
    return in_channels, out_channels


def get_embedding(rule: Rule, arity=1) -> Optional[Tuple[Embedding, str]]:
    if len(rule.body) != 1 or len(rule.body[0].terms) != 1 or rule.head.predicate.arity != arity:
        return None

    in_channels, out_channels = get_in_out_from_atom(rule.head)

    module = Embedding(num_embeddings=out_channels, embedding_dim=in_channels, name=rule.head.predicate.name)
    module.features_name = rule.body[0].predicate.name

    return module, str(rule.body[0].predicate)


def get_gcn(rule: Rule) -> Optional[Tuple[GCNConv, str]]:
    if len(rule.body) != 2 or rule.head.predicate.arity != 1:
        return None

    [term] = rule.head.terms
    in_channels, out_channels = get_in_out_from_atom(rule.head)

    if rule.body[0].predicate.arity == 1:
        if rule.body[1].predicate.arity != 2:
            return None
        edge_atom = rule.body[1]
        feature_atom = rule.body[0]
    elif rule.body[0].predicate.arity == 2:
        if rule.body[1].predicate.arity != 1:
            return None
        edge_atom = rule.body[0]
        feature_atom = rule.body[1]
    else:
        return None

    if term not in edge_atom.terms:
        return None
    if feature_atom.terms[0] not in edge_atom.terms:
        return None

    module = GCNConv(in_channels=in_channels, out_channels=out_channels, name=rule.head.predicate.name)
    module.edge_name = edge_atom.predicate.name
    module.features_name = feature_atom.predicate.name

    return module, str(feature_atom.predicate)


def get_sage(rules: List[Rule]) -> Optional[Tuple[SAGEConv, str]]:
    if len(rules) != 2:
        return None

    gcn = None
    embedding = None

    for rule in rules:
        if gcn is None:
            gcn = get_gcn(rule)
        if embedding is None:
            embedding = get_embedding(rule)
        if gcn is not None and embedding is not None:
            break
    else:
        return None

    if embedding[1] != gcn[1]:
        return None

    module = SAGEConv(in_channels=gcn[0].in_channels, out_channels=gcn[0].out_channels, name=embedding[0].name)
    module.edge_name = gcn[0].edge_name
    module.features_name = gcn[0].features_name

    return module, gcn[1]


def get_gin(rules: List[Rule]) -> Optional[Tuple[GINConv, str]]:
    sort_by_name = {}

    for rule in rules:
        name = str(rule.head.predicate)

        if name not in sort_by_name:
            sort_by_name[name] = []
        sort_by_name[name].append(rule)

    sage = None
    embedding = None

    for rule_group in sort_by_name.values():
        if sage is None:
            sage = get_sage(rule_group)
        if embedding is None and len(rule_group) == 1:
            embedding = get_embedding(rule_group[0])
        if sage is not None and embedding is not None:
            break
    else:
        return None

    if embedding[1] != f"{sage[0].name}/1":
        return None

    module = GINConv(
        in_channels=embedding[0].num_embeddings, out_channels=embedding[0].embedding_dim, name=embedding[0].name
    )
    module.edge_name = sage[0].edge_name
    module.features_name = sage[0].features_name

    return module, sage[1]


def get_pooling(rules: List[Rule]) -> Optional[Tuple[GlobalPooling, List[str]]]:
    embeddings = []
    last_embedding = None

    if len(rules) == 0:
        return None

    for rule in rules:
        embedding = get_embedding(rule, 0)
        if embedding is None:
            return None
        embeddings.append(embedding[1])
        last_embedding = embedding[0]

    return (
        GlobalPooling(
            in_channels=last_embedding.embedding_dim,
            out_channels=last_embedding.num_embeddings,
            name=last_embedding.name,
        ),
        embeddings,
    )



import os
from typing import Optional

from neuralogic import get_neuralogic
from neuralogic.core.settings import Settings

from py4j.java_gateway import set_field


def get_drawing_settings(img_type: str = "png") -> Settings:
    """Returns the default settings instance for drawing with a specified image type.

    :param img_type:
    :return:
    """
    settings = Settings()

    set_field(settings.settings, "drawing", False)
    set_field(settings.settings, "storeNotShow", True)
    set_field(settings.settings, "imgType", img_type)

    return settings


def get_template_drawer(settings: Settings):
    namespace = get_neuralogic().cz.cvut.fel.ida.pipelines.debugging.drawing

    return namespace.TemplateDrawer(settings.settings)


def get_sample_drawer(settings: Settings):
    namespace = get_neuralogic().cz.cvut.fel.ida.pipelines.debugging.drawing

    return namespace.NeuralNetDrawer(settings.settings)


# todo gusta: + groundingDrawer, pipelineDrawer...


def draw(drawer, obj, filename: Optional[str] = None, draw_ipython=True, *args, **kwargs):
    if filename is not None:
        drawer.drawIntoFile(obj, os.path.abspath(filename))

        return

    data = drawer.drawIntoBytes(obj)

    if draw_ipython:
        from IPython.display import Image

        return Image(data, *args, **kwargs)
    return data


def to_dot_source(drawer, obj) -> str:
    return drawer.getGraphSource(obj)


def draw_model(model, filename: Optional[str] = None, draw_ipython=True, img_type="png", *args, **kwargs):
    """Draws model either as an image of type img_type either into:
        * a file - if filename is specified),
        * an IPython Image - if draw_ipython is True
        * or bytes otherwise

    :param model:
    :param filename:
    :param draw_ipython:
    :param img_type:
    :param args:
    :param kwargs:
    :return:
    """
    if model.need_sync:
        model.sync_template()

    template = model.template
    template_drawer = get_template_drawer(get_drawing_settings(img_type=img_type))

    return draw(template_drawer, template, filename, draw_ipython, *args, **kwargs)


def draw_sample(sample, filename: Optional[str] = None, draw_ipython=True, img_type="png", *args, **kwargs):
    """Draws sample either as an image of type img_type either into:
        * a file - if filename is specified),
        * an IPython Image - if draw_ipython is True
        * or bytes otherwise

    :param sample:
    :param filename:
    :param draw_ipython:
    :param img_type:
    :param args:
    :param kwargs:
    :return:
    """
    sample_drawer = get_sample_drawer(get_drawing_settings(img_type=img_type))

    return draw(sample_drawer, sample, filename, draw_ipython, *args, **kwargs)


def model_to_dot_source(model) -> str:
    """Renders the model into its dot source representation.

    :param model:
    :return:
    """
    if model.need_sync:
        model.sync_template()

    template = model.template
    template_drawer = get_template_drawer(get_drawing_settings())

    return to_dot_source(template_drawer, template)


def sample_to_dot_source(sample) -> str:
    """Renders the sample into its dot source representation.

    :param sample:
    :return:
    """
    sample_drawer = get_sample_drawer(get_drawing_settings())

    return to_dot_source(sample_drawer, sample)



from typing import List, Iterable

from neuralogic.core import Template, Relation
from neuralogic.utils.templates.modules import AbstractModule

from neuralogic.utils.templates.modules.gcn import GCNConv
from neuralogic.utils.templates.modules.gsage import SAGEConv
from neuralogic.utils.templates.modules.gin import GINConv
from neuralogic.utils.templates.modules.pooling import GlobalPooling
from neuralogic.utils.templates.modules.embedding import Embedding


class TemplateList:
    """TemplateList is a collection of pre-defined modules"""
    edge_name = "edge"
    feature_name = "node_feature"
    output_name = "predict"

    def __init__(self, modules: List[AbstractModule]):
        self.modules = modules

    def build(self, template: Template):
        if len(self.modules) == 0:
            return

        with template.context():
            previous_names = []

            for i, component in enumerate(self.modules):
                previous_names.append(
                    component.build(template, i + 1, previous_names, self.feature_name, self.edge_name)
                )

    @staticmethod
    def to_inputs(template: Template, x: Iterable, edge_index: Iterable, y, y_mask):
        with template.context():
            if y is None:
                queries = [Relation.get(TemplateList.output_name)]
            else:
                queries = [
                    Relation.get(TemplateList.output_name)(*term if isinstance(term, list) else term)[value]
                    for term, value in zip(y_mask, y)
                ]
            example = [
                Relation.get(TemplateList.edge_name)(int(u), int(v))[1] for u, v in zip(edge_index[0], edge_index[1])
            ]

            for i, features in enumerate(x):
                example.append(Relation.get(TemplateList.feature_name)(i)[features])
        return queries, example



from neuralogic.core.constructs.factories import Var, Term, Relation
from neuralogic.core.template import Template, BuiltDataset
from neuralogic.core.constructs.java_objects import JavaFactory
from neuralogic.core.constructs.metadata import Metadata
from neuralogic.core.settings import Settings, Activation, Aggregation, ErrorFunction, Optimizer, Initializer
from neuralogic.core.builder import Backend
